import { useRef, useCallback, useEffect, useState } from 'react';

/**
 * This hook behaves like a regular useEffect except that it provides a timeout function via it's arguments
 * that can be used to spawn timeouts via `window.setTimeout()` from inside the effect. The timeouts will be cleared on unmount.
 * @param effect The effect callback, it receives a timeout function as its first argument and a timeout clearer as its second
 * @param deps The dependency array from useEffect.
 */
const useTimeoutEffect = (effect, deps) => {
    const timeoutIds = useRef([]);
    const timeoutFunc = useCallback((handler, timeout) => {
        const id = setTimeout(handler, timeout);
        timeoutIds.current.push(id);
        return id;
    }, [timeoutIds]);
    useEffect(() => {
        return effect(timeoutFunc, () => {
            if (timeoutIds.current.length > 0) {
                timeoutIds.current.forEach((id) => {
                    clearTimeout(id);
                });
            }
        });
    }, deps);
    useEffect(() => {
        return function onUnmount() {
            if (timeoutIds.current.length > 0) {
                timeoutIds.current.forEach((id) => {
                    clearTimeout(id);
                });
            }
        };
    }, [timeoutIds]);
};

/**
 * This hook will return a function that executes the provided callback after the specified amount of time.
 *
 * This **will not debounce or throttle** the callbacks, i.e. consecutive calls of this function will all spawn
 * new timeouts even if some are still pending. If you want a debouncing version, take a look at `useDebounce()`.
 * If you want a throttling version, see `useThrottle()`.
 *
 * Pending callbacks will only(!) be cleared in case the component unmounts.
 *
 * @param callback The callback that is invoked after the timeout expired
 * @param waitMs The timeout in milliseconds
 *
 * @returns a function that executes the provided callback after the specified amount of time
 */
function useTimeout(callback, waitMs) {
    const timeoutCallback = useRef(callback);
    const timeoutIds = useRef([]);
    useEffect(() => {
        timeoutCallback.current = callback;
    }, [callback]);
    useEffect(() => {
        return () => {
            timeoutIds.current.forEach((id) => clearTimeout(id));
        };
    }, [timeoutIds]);
    return useCallback((...args) => {
        const id = setTimeout(() => timeoutCallback.current(...args), waitMs);
        timeoutIds.current.push(id);
        return id;
    }, [waitMs]);
}

/******************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */
/* global Reflect, Promise, SuppressedError, Symbol, Iterator */


function __rest(s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
}

typeof SuppressedError === "function" ? SuppressedError : function (error, suppressed, message) {
    var e = new Error(message);
    return e.name = "SuppressedError", e.error = error, e.suppressed = suppressed, e;
};

/**
 * This private API hook adds controls like play, pause, start, stop to a hook.
 * @param isPausedInitially
 * @param isStoppedInitially
 */
const useControls = (isPausedInitially = false, isStoppedInitially = false) => {
    const [isPaused, setIsPaused] = useState(isPausedInitially);
    const [isStopped, setIsStopped] = useState(isStoppedInitially);
    const isPausedRef = useRef(isPausedInitially);
    const pause = useCallback(() => {
        isPausedRef.current = true;
        setIsPaused(true); // notify interval owner
    }, [setIsPaused]);
    const resume = useCallback(() => {
        isPausedRef.current = false;
        setIsPaused(false); // notify interval owner
    }, [setIsPaused]);
    const stop = useCallback(() => {
        setIsStopped(true);
    }, [setIsStopped]);
    const start = useCallback(() => {
        isPausedRef.current = false;
        setIsPaused(false);
        setIsStopped(false);
    }, [setIsStopped]);
    return {
        isPaused,
        isPausedRef,
        isStopped,
        pause,
        resume,
        stop,
        start,
    };
};

/**
 * Calls the given function at a regular interval.
 *
 * The interval can be paused, resumed, stopped etc. via the returned callbacks.
 *
 * Active intervals will be cleared in case the component unmounts.
 *
 * @param callback A function that will be called at the specified interval
 * @param delay time in milliseconds between each invocation of callback. If set to `null` the interval will come to a halt.
 * @param options A set of options to control the interval behaviour.
 *
 * @returns An object of properties to control the interval or see it's status
 */
const useInterval = (callback, delay, options = {}) => {
    const { startOnMount = false, isLeading = false } = options;
    const _a = useControls(false, !startOnMount), { isPausedRef } = _a, controls = __rest(_a, ["isPausedRef"]);
    const { isStopped } = controls;
    const intervalCallback = useRef(callback);
    const intervalId = useRef(null);
    useEffect(() => {
        intervalCallback.current = callback;
    }, [callback]);
    const onIntervalStep = useCallback(() => {
        if (isPausedRef.current === false) {
            intervalCallback.current();
        }
    }, [intervalCallback, isPausedRef]);
    useEffect(() => {
        if (delay !== null && !isStopped) {
            if (isLeading) {
                onIntervalStep();
            }
            intervalId.current = setInterval(onIntervalStep, delay);
            return () => clearInterval(intervalId.current);
        }
    }, [delay, isStopped]);
    return Object.assign({}, controls);
};

/**
 * A hook that updates a number on a regular interval based on the provided settings.
 *
 * Active intervals will be cleared on unmount.
 *
 * @param settings Counter settings and interval options
 * @param settings.start the initial value
 * @param settings.interval duration in ms between steps
 * @param settings.stepSize amount that is added to the current counter value on every step
 * @param [settings.resetOnStop = true] If true, the counter will reset to the start value on stop. If false, it won't.
 * @param [settings.destroyIntervalOnPause = true] If true, the interval in the background will be destroyed upon pause. If false, it won't.
 * @param [settings.startOnMount = false] If true, the interval will immediately start on mount. If false, it has to be started manually via `start()`.
 */
const useCounter = (settings) => {
    const { start, interval, stepSize, resetOnStop = true, destroyIntervalOnPause = true } = settings, intervalOptions = __rest(settings, ["start", "interval", "stepSize", "resetOnStop", "destroyIntervalOnPause"]);
    const [val, setVal] = useState(start);
    const intervalControls = useInterval(() => setVal(val + stepSize), interval, intervalOptions);
    const reset = useCallback(() => {
        setVal(start);
    }, [setVal]);
    const stopFunction = useCallback(() => {
        if (resetOnStop) {
            reset();
        }
        intervalControls.stop();
    }, [resetOnStop]);
    const counterControls = Object.assign(Object.assign({}, intervalControls), { stop: stopFunction, pause: destroyIntervalOnPause
            ? intervalControls.stop
            : intervalControls.pause, resume: destroyIntervalOnPause
            ? intervalControls.start
            : intervalControls.resume, reset });
    return [val, counterControls];
};

/**
 * A hook that starts a timer, i.e. a reactive number that is increased every second.
 * @param start Starting value of the timer
 * @param settings A set of timer settings.
 * @param [settings.resetOnStop = true] If true, the timer will reset to the start value on stop. If false, it won't.
 * @param [settings.destroyIntervalOnPause = true] If true, the interval in the background will be destroyed upon pause. If false, it won't.
 * @param [settings.startOnMount = false] If true, the interval will immediately start on mount. If false, it has to be started manually via `start()`.
 * @returns the current timer value.
 */
const useTimer = (start = 0, settings = {}) => useCounter(Object.assign({ start, interval: 1000, stepSize: 1 }, settings));

const LIB_NAME = 'React Timing Hooks';
const buildMessage = (message) => {
    return `${LIB_NAME}: ${message}`;
};
const log = (logFunction, message) => {
    var _a;
    if (((_a = process.env) === null || _a === void 0 ? void 0 : _a.NODE_ENV) === 'test') {
        return;
    }
    console[logFunction](buildMessage(message));
};
const logError = (message) => log('error', message);
const logWarning = (message) => log('warn', message);

/**
 * Behaves like a regular use effect except that its callback receives a function-argument that allows
 * to run callbacks via `window.requestIdleCallback` inside the effect. Pending idle callbacks will be cleared on unmount.
 * @param effect
 * @param deps
 */
const useIdleCallbackEffect = (effect, deps) => {
    if (!window.requestIdleCallback) {
        logWarning('This browser does not support "requestIdleCallback"');
        return;
    }
    const idleCallbackHandle = useRef(null);
    const idleCallbackFunc = useCallback((callback, options) => {
        idleCallbackHandle.current = window.requestIdleCallback(callback, options);
        return idleCallbackHandle.current;
    }, [idleCallbackHandle]);
    useEffect(() => {
        return effect(idleCallbackFunc);
    }, deps);
    useEffect(() => {
        return function onUnmount() {
            if (idleCallbackHandle.current !== null) {
                window.cancelIdleCallback(idleCallbackHandle.current);
            }
        };
    }, [idleCallbackHandle]);
};

/**
 * Returns a function that can be used to invoke the given callback via `window.requestIdleCallback()`.
 *
 * Pending callbacks will be cleared in case the component unmounts.
 *
 * @param callback The callback that is invoked as soons as the browser invokes the idle callback
 * @param options Options for requestIdleCallback
 */
const useIdleCallback = (callback, options) => {
    if (!window.requestIdleCallback) {
        logWarning('This browser does not support "requestIdleCallback"');
        return callback;
    }
    const ricCallback = useRef(callback);
    const [handle, setHandle] = useState(null);
    useEffect(() => {
        ricCallback.current = callback;
    }, [callback]);
    useEffect(() => {
        return () => {
            if (handle) {
                window.cancelIdleCallback(handle);
            }
        };
    }, [handle]);
    return useCallback((...args) => {
        const h = window.requestIdleCallback(() => ricCallback.current(...args), options);
        setHandle(h);
    }, [options]);
};

/**
 * Returns a function that can be used to run the given callback via `window.requestAnimationFrame()`.
 *
 * Pending callbacks will be cleared in case the component unmounts.
 *
 * @param callback The callback that is invoked in the next animation frame
 */
const useAnimationFrame = (callback) => {
    const rafCallback = useRef(callback);
    const handleRef = useRef(null);
    useEffect(() => {
        rafCallback.current = callback;
    }, [callback]);
    useEffect(() => {
        return () => {
            if (handleRef.current) {
                cancelAnimationFrame(handleRef.current);
            }
        };
    }, []);
    return useCallback((...args) => {
        handleRef.current = requestAnimationFrame(() => rafCallback.current(...args));
        return handleRef.current;
    }, []);
};

const useAnimationFrameLoop = (callback, options = {}) => {
    const { startOnMount = false } = options;
    const rafCallback = useRef(callback);
    const _a = useControls(false, !startOnMount), { isPausedRef } = _a, controls = __rest(_a, ["isPausedRef"]);
    const { isStopped } = controls;
    useEffect(() => {
        rafCallback.current = callback;
    }, [callback]);
    const nextCallback = useCallback(() => {
        if (!isStopped) {
            if (!isPausedRef.current) {
                rafCallback.current();
            }
            runInLoop();
        }
    }, [isStopped]);
    const runInLoop = useAnimationFrame(nextCallback);
    useEffect(() => {
        if (!isStopped) {
            const h = runInLoop();
            return () => {
                cancelAnimationFrame(h);
            };
        }
    }, [runInLoop, isStopped]);
    return controls;
};

/**
 * Creates a sort of clock, i.e. a reactive time-based value that updates every second.
 * useClock is generic (by default useClock<string> is used). The generic type defines the return type which can be
 * changed by using a custom formatter (see options.customFormatter).
 *
 * @template [T=string]
 * @param options options.locales and options.dateTimeFormatOptions will be directly forwarded to date.toLocaleTimeString().
 *                You can also use options.customFormatter to override the output of the hook. The output must match the generic type of the hook.
 *                options.keepPausedClockRunningInBackground will allow to only pause the output but not the underlying clock, so that it resumes at the correct time.
 * @returns The current (formatted) time
 */
const useClock = (options) => {
    const { keepPausedClockRunningInBackground = true, startTimeInMilliseconds = Date.now(), locales, dateTimeFormatOptions, customFormatter, } = options || {};
    const startTimeInSeconds = startTimeInMilliseconds / 1000;
    const [timeDuringPause, setTimeDuringPause] = useState(null);
    const [currentTimeInSeconds, controls] = useTimer(startTimeInSeconds, {
        startOnMount: true,
        resetOnStop: false, // would be weird UX for a clock
        destroyIntervalOnPause: false, // for better accuracy on pause
    });
    const date = new Date(currentTimeInSeconds * 1000);
    const formattedTime = customFormatter
        ? customFormatter(date)
        : date.toLocaleTimeString(locales, dateTimeFormatOptions);
    const customPause = useCallback(() => {
        if (keepPausedClockRunningInBackground) {
            setTimeDuringPause(formattedTime);
        }
        else {
            controls.pause();
        }
    }, [formattedTime, controls.pause]);
    const customResume = useCallback(() => {
        if (keepPausedClockRunningInBackground) {
            setTimeDuringPause(null);
        }
        else {
            controls.resume();
        }
    }, [controls.resume]);
    const clockControls = Object.assign(Object.assign({}, controls), { pause: customPause, resume: customResume });
    return [timeDuringPause || formattedTime, clockControls];
};

/**
 * A hook creates an oscillating boolean value,
 * i.e. a value that automatically toggles repeatedly (infinitely switching from true to false and back) at a fixed interval.
 *
 * Active intervals will be cleared on unmount.
 *
 * @param settings Oscillator settings and interval options
 */
const useOscillator = (settings) => {
    const { initialValue, interval, resetOnStop = true, destroyIntervalOnPause = true } = settings, intervalOptions = __rest(settings, ["initialValue", "interval", "resetOnStop", "destroyIntervalOnPause"]);
    const [val, setVal] = useState(initialValue);
    const intervalControls = useInterval(() => setVal(!val), interval, intervalOptions);
    const reset = useCallback(() => {
        setVal(initialValue);
    }, [setVal]);
    const stopFunction = useCallback(() => {
        if (resetOnStop) {
            reset();
        }
        intervalControls.stop();
    }, [resetOnStop]);
    const oscillatorControls = Object.assign(Object.assign({}, intervalControls), { stop: stopFunction, pause: destroyIntervalOnPause
            ? intervalControls.stop
            : intervalControls.pause, resume: destroyIntervalOnPause
            ? intervalControls.start
            : intervalControls.resume, reset });
    return [val, oscillatorControls];
};

/**
 * This hook creates a countdown that starts and ends at the specified numbers.
 * By default, the value will be changed every second and count downwards. Both properties can be changed, though.
 *
 * The user will be notified when the countdown ends via the event callback `settings.onEnd()`.
 *
 * Use the returned `start()` callback or set `settings.startOnMount` to start the countdown.
 *
 * @param from Start value
 * @param to End value. Will trigger the `settings.onEnd()` callback.
 * @param settings A set of countdown settings.
 * @param settings.onEnd If set, this callback will be called when the 2nd param "`to`" is reached.
 * @param [settings.resetOnStop = false] If true, the countdown will reset to the start value on stop. If false, it won't.
 * @param [settings.destroyIntervalOnPause = true] If true, the interval in the background will be destroyed upon pause. If false, it won't.
 * @param [settings.startOnMount = false] If true, the interval will immediately start on mount. If false, it has to be started manually via `start()`.
 * @returns an array: the first element is the countdown value, the second is an object of interval controls, see useInterval()
 */
const useCountdown = (from, to, settings = {}) => {
    const onEndCallback = useRef(settings.onEnd);
    const [value, counterControls] = useCounter(Object.assign({ start: from, interval: 1000, stepSize: -1, resetOnStop: false }, settings));
    useEffect(() => {
        onEndCallback.current = settings.onEnd;
    }, [settings.onEnd]);
    useEffect(() => {
        if (to > from &&
            (typeof settings.stepSize === 'undefined' || settings.stepSize < 0)) {
            logError(`Stopped countdown because a countdown from ${from} to ${to} will never end`);
            counterControls.stop();
        }
    }, [from, to]);
    useEffect(() => {
        var _a;
        if (value === to) {
            (_a = onEndCallback.current) === null || _a === void 0 ? void 0 : _a.call(onEndCallback);
            counterControls.stop();
        }
    }, [value]);
    return [value, counterControls];
};

/**
 * Debounces a callback. A debounce is similar to a throttle, but every attempt to call the function resets the
 * block-timer until the function can be executed again. This means that if the debounced callback is consistently
 * called too frequently, it will not be called again at all - until it stops for long enough time.
 *
 * By default, `options.trailing = true`, meaning that the callback will only be invoked as soon as a certain time
 * has passed since the last call (defined by `waitMs`).
 *
 * Alternatively, the function can also be called immediately and then be blocked on consecutive calls until the timeout
 * is over (when `options.leading` is true). You can also set both `options.leading` and `options.trailing` to `true`.
 *
 * @param callback
 * @param waitMs The minimum time until an invocation can happen.
 * @param options
 * @param [options.leading = false] If true, invoke before the timeout
 * @param [options.trailing = true] If true, invoke after the timeout
 */
function useDebounce(callback, waitMs, options = {}) {
    const { leading = false, trailing = true } = options;
    const debouncedCallback = useRef(callback);
    const timeoutId = useRef(null);
    useEffect(() => {
        debouncedCallback.current = callback;
    }, [callback]);
    useEffect(() => {
        return () => {
            if (timeoutId.current) {
                clearTimeout(timeoutId.current);
            }
        };
    }, []);
    return useCallback((...args) => {
        if (timeoutId.current) {
            clearTimeout(timeoutId.current);
        }
        else if (leading) {
            debouncedCallback.current(...args);
        }
        timeoutId.current = setTimeout(() => {
            if (trailing) {
                debouncedCallback.current(...args);
            }
            timeoutId.current = null;
        }, waitMs);
        return timeoutId.current;
    }, [waitMs]);
}

/**
 * Throttles a callback.
 *
 * Can be used for **rate-limiting** – the callback will only be invoked every X milliseconds (X being the set timeout),
 * even if it was called more frequently.
 *
 * Similar, but different(!), is the `useDebounce()` hook, which blocks the invocation entirely until the function was
 * stopped being called for X milliseconds.
 *
 * By default, the throttled function will always be called immediately (`options.leading` is true by default) and then
 * (`options.trailing` is true by default) also after every X milliseconds for consecutive calls.
 *
 * @param callback
 * @param waitMs Minimum waiting time between consecutive calls
 * @param options
 * @param [options.leading = true] If true, invoke the callback immediately/before the timeout
 * @param [options.trailing = true] If true, queue invocations to be called after the timeout
 */
function useThrottle(callback, waitMs, options = {}) {
    const { leading = true, trailing = true } = options;
    const hasQueuedCall = useRef(false);
    const queuedArgs = useRef(null);
    const throttledCallback = useRef(callback);
    const timeoutId = useRef(null);
    useEffect(() => {
        throttledCallback.current = callback;
    }, [callback]);
    useEffect(() => {
        return () => {
            if (timeoutId.current) {
                clearTimeout(timeoutId.current);
            }
        };
    }, []);
    const execThrottled = useCallback((...args) => {
        if (timeoutId.current) {
            if (trailing) {
                hasQueuedCall.current = true;
                queuedArgs.current = args;
            }
            return timeoutId.current;
        }
        if (!hasQueuedCall.current) {
            if (leading) {
                throttledCallback.current(...args);
            }
            else if (trailing) {
                hasQueuedCall.current = true;
                queuedArgs.current = args;
            }
        }
        timeoutId.current = setTimeout(() => {
            timeoutId.current = null;
            if (hasQueuedCall.current) {
                if (trailing) {
                    throttledCallback.current(...queuedArgs.current);
                }
                // restart loop with(!) a queued call, but without a timer running, essentially just restarting the timer
                execThrottled(...queuedArgs.current);
                // remove queued call afterwards
                hasQueuedCall.current = false;
            }
        }, waitMs);
        return timeoutId.current;
    }, [waitMs]);
    return execThrottled;
}

/** Behaves like a regular useState hook, except that setting the state
 *  is throttled, i.e. a minimum amount of time will have to pass between
 *  state updates, defined by the specified delay. This allows for rapid
 *  state updates without the actual state value causing too many updates/re-renders. */
function useThrottledState(initialState, delayInMs) {
    const [actualStateValue, setActualStateValue] = useState(initialState);
    const setState = useThrottle(setActualStateValue, delayInMs);
    return [actualStateValue, setState];
}

export { useAnimationFrame, useAnimationFrameLoop, useClock, useCountdown, useCounter, useDebounce, useIdleCallback, useIdleCallbackEffect, useInterval, useOscillator, useThrottle, useThrottledState, useTimeout, useTimeoutEffect, useTimer };
